extends KinematicBody

const physics_speed = 1
const GRAVITY = -120 * physics_speed
const JUMP_SPEED = 45 * physics_speed
const PLAYER_SPEED = 30 * physics_speed
const MAX_SLOPE_ANGLE = deg2rad(60)
const CAMERA_CLAMP_ANGLE = deg2rad(89)
const MAX_SLIDES = 4

export(float, 0.1, 100.0) var camera_distance = 15.0
export var follow_floor_rotation: bool = true
var mouse_sensitivity = 0.01

onready var player = $"."
onready var player_body = $PlayerBody
onready var model = $PlayerBody
onready var camera_helper = $CameraHelper
onready var camera = $CameraHelper/CameraPosition/Camera
onready var camera_position = $CameraHelper/CameraPosition
onready var debug_info = $DebugInfo

const pos_marker = preload("res://PosMarker.tscn")

var dir: Vector3
var input_movement_vector: Vector2
var vel: Vector3 = Vector3()
var camera_rotation = Vector3()
var gravity_vel: Vector3 = Vector3()

export var use_improved_approximation: bool = true

func _ready():
	Input.set_mouse_mode(Input.MOUSE_MODE_CAPTURED)
	camera_position.translate_object_local(Vector3(0, 0, camera_distance))
	debug_info.add("fps", 0)
	debug_info.add("time", 0)
	debug_info.add("use_improved_approximation", use_improved_approximation)

func _process(_delta):
	debug_info.add("fps", Engine.get_frames_per_second())
	debug_info.add("time", OS.get_ticks_msec() / 1000.0)
	debug_info.add("use_improved_approximation", str(use_improved_approximation) + " (Press Q to toggle)" )
	debug_info.render()

func _physics_process(delta):
	process_input(delta)
	process_movement(delta)
	add_position_marker(delta)

func improved_floor_velocity_estimate(delta):
	"""
	This function returns an estimate of the floor velocity for use in
	move_and_slide(). The result should be added to the `velocity` argument of
	`move_and_slide()`.
	
	NOTE: since `move_and_slide()` adds `get_floor_velocity()` internally,
	this function includes `-get_floor_velocity()` in the result to cancel this
	out.
	"""
	if !is_on_floor():
		debug_info.add("angular_velocity", null)
		return Vector3()
	
	return -get_floor_velocity() + floor_angular_velocity(delta)

func floor_angular_velocity(delta):
	# FIXME: don't assume that the first object we collided with was the floor 
	var collision = get_slide_collision(0)
	var floor_node: Node = collision.collider
	var ang_vel = PhysicsServer.body_get_direct_state(floor_node.get_rid()).angular_velocity
	
	if ang_vel.length_squared() == 0:
		return Vector3()
	
	# the origin point in global coordinates
	var rotation_origin = floor_node.global_transform.origin
	# NOTE: collision.point is out of date compared to our current position
	# so want to update it for this frame
	# see also: https://github.com/godotengine/godot-proposals/issues/2332
	# TODO: work this out better
	var current_collision_pos = self.global_transform.origin
	var collision_pos_relative = current_collision_pos - rotation_origin
	var next_rotated_xform = Transform().rotated(ang_vel.normalized(), ang_vel.length()*delta)
	var collision_pos_relative_next = next_rotated_xform.xform(collision_pos_relative)
	
	var v_avg = (collision_pos_relative_next - collision_pos_relative) / delta

	debug_info.add("angular_velocity", ang_vel)
	return v_avg

func compute_player_dir():
	dir = Vector3.ZERO

	if input_movement_vector != Vector2():
		var cam_transform = camera.get_global_transform()
		dir += cam_transform.basis.x * input_movement_vector[0]
		dir += -cam_transform.basis.z * input_movement_vector[1]
		dir.y = 0
		dir = dir.normalized()


func process_movement(delta):
	compute_player_dir()

	var player_vel = dir * PLAYER_SPEED
	gravity_vel += Vector3.UP * delta * GRAVITY
	
	var floor_vel_adjust 
	if use_improved_approximation:
		floor_vel_adjust = improved_floor_velocity_estimate(delta)
	else:
		floor_vel_adjust = Vector3()
		
	vel = player_vel + gravity_vel + floor_vel_adjust
	vel = player.move_and_slide(vel, Vector3.UP, false, MAX_SLIDES)
	
	if is_on_floor():
		gravity_vel = Vector3()
	
	debug_info.plot_bool("is_on_floor", is_on_floor())
	debug_info.add("floor_velocity", get_floor_velocity())


func process_input(_delta):
	if Input.is_action_just_pressed("ui_cancel"):
		if Input.get_mouse_mode() == Input.MOUSE_MODE_VISIBLE:
			Input.set_mouse_mode(Input.MOUSE_MODE_CAPTURED)
		else:
			Input.set_mouse_mode(Input.MOUSE_MODE_VISIBLE)
	
	if Input.is_action_just_pressed("toggle_vel_func"):
		use_improved_approximation = !use_improved_approximation
			
	input_movement_vector = Vector2()
	if Input.is_action_pressed("movement_forward"):
		input_movement_vector.y += 1
	if Input.is_action_pressed("movement_backward"):
		input_movement_vector.y -= 1
	if Input.is_action_pressed("movement_left"):
		input_movement_vector.x -= 1
	if Input.is_action_pressed("movement_right"):
		input_movement_vector.x += 1

	input_movement_vector = input_movement_vector.normalized()

	if is_on_floor() and Input.is_action_just_pressed("movement_jump"):
		gravity_vel.y = JUMP_SPEED

var marker_timer: float = 0.0
func add_position_marker(delta):
	marker_timer += delta
	if marker_timer > 0.1:
		var new_marker = pos_marker.instance()
		new_marker.transform = self.transform
		$"..".add_child(new_marker)
		marker_timer = 0.0

func _input(event):
	if event is InputEventMouseMotion and Input.get_mouse_mode() == Input.MOUSE_MODE_CAPTURED:
		var yaw = camera_rotation.y + event.relative.x * mouse_sensitivity * -1
		var pitch = camera_rotation.x + -event.relative.y * mouse_sensitivity
		yaw = fmod(yaw, 2*PI)
		pitch = clamp(pitch, -CAMERA_CLAMP_ANGLE, CAMERA_CLAMP_ANGLE)
		camera_rotation = Vector3(pitch, yaw, 0)
		
		camera_helper.rotation = camera_rotation
